"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.
"""

from .odict import ODict
import six
import yaml

TAG_MAP = "tag:yaml.org,2002:map"
TAG_STRING = "tag:yaml.org,2002:str"


class CfnYamlDumper(yaml.Dumper):
    """
    Indent block sequences from parent using more common style
    ("  - entry"  vs "- entry").
    Causes fewer problems with validation and tools.
    """

    def increase_indent(self, flow=False, indentless=False):
            return super(CfnYamlDumper, self).increase_indent(flow, False)

    def represent_scalar(self, tag, value, style=None):
            if isinstance(value, six.text_type):
                if "\n" in value and style is None:
                    style = "\""

                # return super(CfnYamlDumper, self).represent_scalar(TAG_STRING, value, style)

            return super(CfnYamlDumper, self).represent_scalar(tag, value, style)


def string_representer(dumper, value):
    style = None

    if "\n" in value:
        style = "\""

    return dumper.represent_scalar(TAG_STRING, value, style=style)


def map_representer(dumper, value):
    """
    Map ODict into ODict to prevent sorting
    """

    return dumper.represent_mapping(TAG_MAP, value)


# Customise the dumper
CfnYamlDumper.add_representer(ODict, map_representer)
CfnYamlDumper.add_representer(six.text_type, string_representer)



"""
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from .yaml_dumper import get_dumper
from cfn_clean import clean
from cfn_tools import load_json, load_yaml, dump_json
import yaml


def load(template):
    """
    Try to guess the input format
    """

    try:
        data = load_json(template)
        return data, "json"
    except ValueError:
        data = load_yaml(template)
        return data, "yaml"


def dump_yaml(data, clean_up=False, long_form=False):
    """
    Output some YAML
    """

    return yaml.dump(
        data,
        Dumper=get_dumper(clean_up, long_form),
        default_flow_style=False
    )


def to_json(template, clean_up=False):
    """
    Assume the input is YAML and convert to JSON
    """

    data = load_yaml(template)

    if clean_up:
        data = clean(data)

    return dump_json(data)


def to_yaml(template, clean_up=False, long_form=False):
    """
    Assume the input is JSON and convert to YAML
    """

    data = load_json(template)

    if clean_up:
        data = clean(data)

    return dump_yaml(data, clean_up, long_form)


def flip(template, out_format=None, clean_up=False, no_flip=False, long_form=False):
    """
    Figure out the input format and convert the data to the opposing output format
    """

    data = None
    in_format = None

    if no_flip:
        in_format = out_format
    elif out_format == "json":
        in_format = "yaml"
    elif out_format == "yaml":
        in_format = "json"

    if in_format == "json":
        data = load_json(template)
    elif in_format == "yaml":
        data = load_yaml(template)
    else:
        try:
            data, in_format = load(template)
        except Exception:
            raise Exception("Could not determine the input format")

    if no_flip:
        out_format = in_format
    elif in_format == "json":
        out_format = "yaml"
    else:
        out_format = "json"

    if clean_up:
        data = clean(data)

    if out_format == "json":
        return dump_json(data)

    return dump_yaml(data, clean_up, long_form)



"""
Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License").
You may not use this file except in compliance with the License. A copy of the License is located at

    http://aws.amazon.com/apache2.0/

or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations under the License.
"""

from cfn_tools.odict import ODict
import six


def convert_join(value):
    """
    Fix a Join ;)
    """

    if not isinstance(value, list) or len(value) != 2:
        # Cowardly refuse
        return value

    sep, parts = value[0], value[1]

    if isinstance(parts, six.string_types):
        return parts

    if not isinstance(parts, list):
        # This looks tricky, just return the join as it was
        return {
            "Fn::Join": value,
        }

    plain_string = True

    args = ODict()
    new_parts = []

    for part in parts:
        part = clean(part)

        if isinstance(part, dict):
            plain_string = False

            if "Ref" in part:
                new_parts.append("${{{}}}".format(part["Ref"]))
            elif "Fn::GetAtt" in part:
                params = part["Fn::GetAtt"]
                new_parts.append("${{{}}}".format(".".join(params)))
            else:
                for key, val in args.items():
                    if val == part:
                        param_name = key
                        break
                else:
                    param_name = "Param{}".format(len(args) + 1)
                    args[param_name] = part

                new_parts.append("${{{}}}".format(param_name))

        elif isinstance(part, six.string_types):
            new_parts.append(part.replace("${", "${!"))

        else:
            # Doing something weird; refuse
            return {
                "Fn::Join": value
            }

    source = sep.join(new_parts)

    if plain_string:
        return source

    if args:
        return ODict((
            ("Fn::Sub", [source, args]),
        ))

    return ODict((
        ("Fn::Sub", source),
    ))


def clean(source):
    """
    Clean up the source:
    * Replace use of Fn::Join with Fn::Sub
    """

    if isinstance(source, dict):
        for key, value in source.items():
            if key == "Fn::Join":
                return convert_join(value)
            else:
                source[key] = clean(value)

    elif isinstance(source, list):
        return [clean(item) for item in source]

    return source
